# DEPRECIATED
# getScreenResolution (xdpyinfo | grep dimensions)
# verifyInput & collectInputVector: use 'menu' instead.

#' Shorthand for \code{x[x \%in\% y]}.
#'
#' @param x vector to test
#' @param y vector to check membership of
#' @param inv whether we are checking membership (\code{\%in\%}) or
#'            exclusion (\code{\%notin\%})
#' @param value whether to return the values of `x` that are in (or not) `y`,
#'        or the indices. If `value=F` this is just like \code{\%in\%} or \code{\%notin\%}.
#' @return If `value=T`, the subset of `x` that is in `y` (`inv=F`), or that is
#'         not in y (`inv=T`).
#'         If `value=F`, a logical vector that `x \%in\% y` if `inv=F` or
#'         `!(x \%in\% y)` otherwise.
#' @seealso \code{\link[base:match]{\%in\%}}
#' @seealso \code{\link[utilitiesR:notin]{\%notin\%}}
#' @seealso \code{\link{invwhich}}
#' @examples
#' x <- 1:5
#' y <- sample(x, size=3)
#' find.in(x, y) # all elements of x that are in y
#' find.in(x, y, inv=TRUE) # all elements of x that are not in y
#'
#' find.in(x, y, value=FALSE) # equivalent to x %in% y
#' find.in(x, y, value=FALSE, inv=TRUE) # equivalent to !(x %in% y)
#' @export
find.in <- function (x, y, inv=FALSE, value=TRUE) {
    idx <- x %in% y
    if (inv)
        idx <- !idx
    if (value) {
        x[idx]
    } else {
        idx
    }
}

#' Opposite to \code{\%in\%}.
#'
#' Returns a logical vector for elements of `x` not in `table`.
#' It is simply equivalent to \code{!(x \%in\% table)}.
#' @usage x \%notin\% table
#' @param x     vector to test
#' @param table vector to check membership of
#' @return a logical vector, same length as `x`, that is TRUE in the `i`th position
#' if `x[i]` is NOT in `table`.
#' @examples
#' x <- 1:5
#' x[x %notin% c(1, 3, 5)] # 2, 4
#' @export
#' @seealso \code{\link[base:match]{\%in\%}}
#' @seealso \code{\link{find.in}}
#' @seealso \code{\link{invwhich}}
#' @aliases %notin%
#' @rdname notin
#' @export
"%notin%" <- function (x, table) {
    match(x, table, nomatch=0L) == 0L
}

#' Opposite to \code{which}.
#'
#' @param indices the indices that should be TRUE
#' @param length.out the length of the output vector. if NULL, `max(indices)` is used.
#' @param use.names whether to use the names of `indices` if there are any.
#' @author user Nick Sabbe on StackOverflow: http://stackoverflow.com/a/7661128/913184
#' @return a logical vector of length `length.out` that is TRUE at the indices
#'         in `indices`.
#' @examples
#' idx <- c(1, 3)
#' invwhich(idx)     # c(TRUE, FALSE, TRUE)
#' invwhich(idx, 5)  # c(TRUE, FALSE, TRUE, FALSE, FALSE)
#'
#' ## example with use.names
#' names(idx) <- c('one', 'three')
#' invwhich(idx, 4)
#' @seealso \code{\link[base:match]{\%in\%}}
#' @seealso \code{\link{find.in}}
#' @seealso \code{\link[utilitiesR:notin]{\%notin\%}}
#' @export
invwhich <- function(indices, length.out=NULL, use.names=TRUE) {
    if (missing(length.out) || is.null(length.out)) {
        length.out <- max(indices)
    }
    rv <- logical(length.out)
    if(length(indices) > 0) {
        rv[indices] <- TRUE
        if (use.names)
            names(rv)[indices] <- names(indices)
    }
    return(rv)
}

#' See if an argument is na, null, or length 0 (syntactic sugar).
#' @param x object to check
#' @return boolean, TRUE if it is na, null OR length 0.
#' @note syntactic sugar for \code{is.null(x) || is.na(x) || length(x)==0}.
#' Note that the empty string ("") is NOT "empty" in this sense.
#' @examples
#' is.empty(NA)          # true
#' is.empty(NULL)        # true
#' is.empty(character()) # true, length 0
#' is.empty("")          # false, this has length 1.
#' @export
is.empty <- function(x)
{
    return(is.null(x) || is.na(x) || length(x)==0)
}

# ******************************** MATHS ******************************** #
#' make 2D rotation matrix anticlockwise by theta
#' @param th amount to rotate by (radians)
#' @return the matrix:
#' \deqn{\left(\begin{array}{cc}
#'      \cos(th) & -\sin(th)\\
#'      \sin(th) & \cos(th)
#'      \end{array}\right)}
#' @note  \code{make2Drot(th) * t(x, y)} gives (x, y) rotated by angle theta
#' anticlockwise.
#' @export
make2Drot <- function(th) {
    return(matrix(c(cos(th), -sin(th),
                     sin(th), cos(th)), ncol=2,  byrow=TRUE))
} # make2Drot

#' return euclidean distance between successive rows in a matrix.
#' @param m    An nxm matrix where there are n points in \eqn{R^m}, and we
#'             want to find the distance between successive rows (points).
#'             i.e., one point per row.
#' @param circular boolean, whether to also calculate the distance between
#' the last and first point.
#' @param cols If this is TRUE, the matrix is mxn instead (each point is a
#'             column)
#' @return a vector of length (n-1) containing the distance between
#'         successive points in "m".
#' @examples
#' # make a matrix of 4 points in 2D.
#' m <- array(runif(4*2), dim=c(4, 2))
#' euclidDist(m)            # dist between rows
#' euclidDist(m, circular=TRUE) # dist between rows including 4 to 1.
#' euclidDist(m, cols=TRUE)     # dist between cols
#' @export
euclidDist <- function(m, cols=FALSE, circular=FALSE) {
    if (cols)
        m <- t(m)
    idx <- 1:nrow(m)
    if (circular)
        idx <- c(idx, 1)
    return(sqrt(rowSums(diff(as.matrix(m[idx, ]))^2)))
}

#' sort vertices in a convex quadrilateral in anticlockwise fashion.
#' @param pts           the input points, nx2, colnames 'X' and 'Y'
#' @param anticlockwise boolean, whether to return points in anticlockwise
#'                      or clockwise fashion (default TRUE)
#' @param shift         default FALSE, whether to return results relative
#'                      to origin
#' @param preserveStart default FALSE, whether to start at the first
#'                      input coordinate and proceed anticlockwise/
#'                      clockwise from there, or to start from the
#'                      lower-left corner ('origin')
#' @param index.return  default FALSE, whether to return the indices into
#'                      the original "pts" that reorder as specified.
#' @return the pts structure with columns 'X', 'Y', and rows sorted in
#'         anticlockwise or clockwise order.
#'         If "index.return" is TRUE, returns a list:
#'         \describe{
#'             \item{x}{the points array}
#'             \item{ix}{indices into "pts" that reorder it into output$x}
#'         }
#'         Additionally, the row-names of the output are labelled
#'          "bottom-left", "bottom-right", "top-left", and "top-right",
#'         based on an X-Y coordinate system, 0 being in the bottom left,
#'          increasing moving left-to-right and bottom-to-top.
#' @details
#' This calculates the coordinates of each vertex relative to the centroid of
#' the points and sorts them by argument, assuming the bottom-left corner to
#' be the one with angle closest to 5*pi/4.
#'
#' Note that if two points are exactly equidistant to 5*pi/4, the one that occurs
#' first in the `pts` array is labelled the 'bottom-left corner' (as it is ambiguous
#' in this case).
#'
#' @note
#' In versions <= 1.8.5, this used a naive method of determining "bottom-left",
#' "top-right", etc. vertices, that would sometimes be WRONG! (particularly
#' in the case of a quadrilateral that was in a diagonal orientation (where
#' diagonally opposite corners had roughly equal Y values).
#'
#' @export
sortPointsRectangle <- function(pts, anticlockwise=TRUE,# verbose=FALSE,
                                shift=FALSE, preserveStart=FALSE, index.return=FALSE) {
    coords <- pts
    sz <- dim(pts)
    if (sz[1] == 2 && sz[2] != 2) {
        coords <- t(pts)
    }
    if (!is.empty(colnames(coords))) {
        colnames(coords) <- toupper(colnames(coords))
        coords <- coords[, c('X', 'Y')]
    }
    n <- nrow(coords)

    # convert to polar space from the centre and then sort.
    centre <- colSums(coords) / n
    # work out the argument of each point with respect to the centroid, in
    # [0, 2pi]. If this is sorted ascending it will be anticlockwise
    thetas <- (atan2(coords[, 2] - centre[2],coords[, 1] - centre[1]) + 2 * pi) %% (2 * pi)

    idx <- order(thetas)

    # work out where bottom-left is.
    rn <- c('bottom-left', 'bottom-right', 'top-right', 'top-left')

    # sort such that we start at the south west corner.
    SWI <- which.min(abs(5 * pi / 4 - thetas))
    idx <- idx[(which(SWI == idx) - 1 + 0:3) %% n + 1]
    names(idx) <- rn

    # now sort back such that we start at sI if that is what the user requested
    # (i'm sure i'm doing this very roundabout-ly...)
    if (preserveStart) {
        idx <- idx[(which(1 == idx) - 1 + 0:3) %% n + 1]
    }

    # reverse the order if they want clockwise.
    if (!anticlockwise) {
        idx <- idx[c(1, n:2)]
    }

    out <- coords[idx, ]
    rownames(out) <- names(idx)

    # shift relative to bottom-left point.
    if (shift) {
        # dumb t(...) because R is col-major
        out <- t(t(out) - out['bottom-left', ])
    }

    if (index.return) {
        return(list(x=out, ix=idx))
    } else {
        return(out)
    }
} # sortPointsRectangle

#' converts GPS coordinates to UTM coordinates
#' @param gps a nx2 matrix with LON then LAT
#' @param using one of 'proj' or 'rgdal': use proj executable (must be in
#'              $PATH) or rgdal package
#' @param inverse whether to invert (ie UTM2GPS), default FALSE
#'               (i.e. GPS -> UTM)
#' @param zone the zone for conversion (either zone the UTM coordinates are in,
#'        or what zone to give UTM coordinates relative to).
#'        If doing GPS to UTM conversion and the zone argument is missing, we
#'         will calculate the zone that the average longitude is in and give all
#'         UTMs in that zone (even if they are infinite!)
#'        If doing GPS to UTM conversion and the zone argument is NULL, we will
#'         give *each* converted UTM in the closest zone (zone includes the
#'         start point and excludes the end point, e.g. Zone 1 is [-180, -174))
#'        In either case, attr(out, 'zone') will be a vector with the zone.
#'        For UTM to GPS conversions, you MUST supply a zone.
#' @param hemisphere hemisphere for conversion.
#'        If doing UTM to GPS conversion, this MUST be provided.
#'        If doing GPS to UTM conversion and this is missing, we provide all UTMs
#'         relative to the hemisphere that the mean latitude is in (0 is North).
#'        If doing GPS to UTM conversion and this is NULL, we provide all UTMs
#'         relative to the hemisphere that each coordinate is in.
#'        In either case, attr(out, 'hemisphere') is a vector with the hemispheres.
#' @return a nx2 matrix with UTM (colnames 'X' and 'Y') or GPS ('lon' and 'lat').
#'         For GPS to UTM conversion:
#'         In addition, it has attributes 'zone' and 'hemisphere' being vectors
#'         of length n with the zone/hemisphere of each coordinate.
#'         Coordinates that could not be converted are set to NA.
#' @details
#' You need either package 'rgdal' [need libgdal1-dev in ubuntu and libproj-dev] OR 'proj' in your path [proj-bin in ubuntu].
#'
#' If doing conversion from UTM to GPS, it is *very important* to
#' know the zone and the hemisphere, as an easting/northing pair can be
#' valid in both hemispheres, and/or multiple zones.
#'
#' If doing conversion from GPS to UTM, it's not as improtant to know the
#' zone/hemisphere, as the hemisphere can be told from the latitude.
#'
#' The zone can be calculated as (wikipedia):
#'
#' The UTM system divides the surface of the Earth ...
#'  into 60 zones, each of 6 degrees of longitude in width.
#' Zone 1 covers longitude 180 to 174 degrees W; zone numbering increases
#' eastward to zone 60 that covers longitude 174 to 180 E.
#'
#' @family coordinates
#' @export
# @importFrom rgdal project
GPS2UTM <- function(gps, using=c('proj', 'rgdal'), inverse=FALSE, zone=56,
                    hemisphere=c('south', 'north')) {
    # work out whether to use rgdal or not
    .USING_RGDAL <- NA
    PROJ <- getProgram('proj') # will do proj.exe too
    if (missing(using)) { # automagically determine what to use
        .USING_RGDAL <- is.null(PROJ)
    } else {
        .USING_RGDAL <- match.arg(using)=='rgdal'
    }

    # check for the required components.
    if (.USING_RGDAL && !require(rgdal))
        stop("You need to install the rgdal package.\n(On Ubuntu, this requires libgdal1-dev and libproj-dev).\nExiting\n")
    if (!.USING_RGDAL && is.null(PROJ))
        stop('You need to install proj: sudo (yum|apt-get|etc) proj (on ubuntu, \'proj-bin\'). If it\'s already installed, then you need to add it to your $PATH.\nAlternatively, install package rgdal to R: install.packages("rgdal")\nExiting.\n')

    n <- nrow(gps)
    if (!inverse) {
        if (missing(zone)) {
            zone <- getUTMZone(mean(gps[, 1]))
        } else if (is.null(zone)) {
            zone <- getUTMZone(gps[, 1])
        }
    } else if (missing(zone)) {
        warning('You must supply a zone for UTM to GPS conversion, we will use ', zone)
    }
    zone <- rep(zone, length.out=n)

    # determine hemisphere in forward direction...check latitude.
    if (!inverse) {
        if (missing(hemisphere)) {
            hemisphere <- ifelse(mean(gps[, 2]) >= 0, 'north', 'south')
        } else if (is.null(hemisphere)) {
            hemisphere <- ifelse(gps[, 2] >= 0, 'north', 'south')
        }
    } else if (missing(hemisphere)) {
        hemisphere <- match.arg(hemisphere)
        warning('You must supply a hemisphere for UTM to GPS conversion, we will use ', hemisphere)
    }
    hemisphere <- rep(hemisphere, length.out=n)

    # now calculate UTMs (or GPS) according to unique (zone, hemisphere) pairs.
    prs <- unique(matrix(c(zone, hemisphere), ncol=2, nrow=n))
    out <- matrix(NA, ncol=ncol(gps), nrow=nrow(gps))
    PROJ <- shQuote(PROJ)
    for (i in 1:nrow(prs)) {
        idx <- which(zone == prs[i, 1] & hemisphere == prs[i, 2])
        out[idx, ] <- utilitiesR:::GPS2UTM.single.zone.hemisphere(gps[idx, ],
                                                                  .USING_RGDAL,
                                                                  inverse,
                                                                  prs[i, 1],
                                                                  prs[i, 2],
                                                                  PROJ)
    }
    if (inverse) {
        colnames(out) <- c("lon", "lat")
    } else {
        colnames(out) <- c("X", "Y")
        attr(out, 'zone') <- zone
        attr(out, 'hemisphere') <- hemisphere
    }
    out[is.infinite(out)] <- NA

    if(any(is.na(out))) {
        warning("Warning: some values were not convertible.")
    }

    return(out)
} # GPS2UTM

#' Internal function. All arguments (bar `gps`) must be scalar. Don't use.
#'
#' Expects rgdal/proj checks have already been loaded/done (will not do existence checking)
#' @inheritParams GPS2UTM
#' @param using.rgdal are we using rgdal?
#' @param PROJ shQuote'd path to proj executable
#' @keywords internal
GPS2UTM.single.zone.hemisphere <- function(gps, using.rgdal, inverse, zone, hemisphere, PROJ) {
    # projection arguments.
    PROJ_ARGS <- sprintf("+proj=utm +datum=WGS84 +zone=%s %s -f '%%f'",
                         as.character(zone),
                         ifelse(hemisphere=='south', '+south', ''))

    if (using.rgdal) {
        utms <- suppressWarnings(rgdal::project(as.matrix(gps), PROJ_ARGS, inv=inverse))
    } else { # use command line `proj`.
        if (inverse) {
            PROJ <- paste(PROJ, '-I')
        }
        if (.Platform$OS == 'windows') {
            # don't know how to make windows do proj .. <<EOF ... EOF
            coordFile <- tempfile()
            write.table(gps[, c(1, 2)], file=coordFile, row.names=FALSE, col.names=FALSE)
            cmd <- sprintf("%s %s %s", PROJ, PROJ_ARGS,
                           shQuote(coordFile))
        } else {
            cmd <- sprintf("%s %s <<EOF
%s
EOF",
                           PROJ,
                           PROJ_ARGS,
                           paste(gps[, 1], gps[, 2], collapse='\n'))
        }
        utmstxt <- system(cmd, intern=TRUE)  # x\ty  or easting\tnorthing
        # Note: on Windows using Cygwin's 'proj.exe' you get a warnin gup the top "MS-DOS style path detected ..."
        # Make sure we ignore lines that do not conform. (but * * for failed conversions count)
        utmstxt <- utmstxt[grepl('^-?[*0-9.]+\t-?[*0-9.]+$', utmstxt)]
        utms <- suppressWarnings(as.numeric(unlist(strsplit(utmstxt, '\t'))))
        utms <- matrix(utms, ncol=2, byrow=TRUE)
    }
    return(utms)
}

#' Calculates the UTM zone a longitude falls in.
#' @param long a vector of longitudes.
#' @return a vector of the same length as `long` with the zone the longitude is in.
#' @details
#' From Wikipedia[1]:
#'
#' The UTM system divides the Earth between 80S and 84N latitude into 60 zones,
#' each 6 degrees of longitude in width. Zone 1 covers longitude 180W to 174W; zone
#' numbering increases eastward to zone 60 that covers longitude 174 to 180E.
#'
#' We take each interval to include the starting longitude and exclude the ending
#' one, so e.g. Zone 1 is [-180, -174); Zone 2 is [-174, -168); ..., Zone 60 is
#' [174, 180).
#' @seealso http://en.wikipedia.org/wiki/Universal_Transverse_Mercator_coordinate_system
#' @family coordinates
#' @export
getUTMZone <- function(long) {
    # 1. convert long to [0, 360]
    long <- (long + 180) %% 360
    # 2. divide into 6-degree intervals
    return(floor(long / 6) + 1)
}

#' converts UTM coordinates to GPS coordinates
#' @param utm a nx2 matrix with X then Y
#' @param ... params for UTM2GPS
#' @return a nx2 matrix with gps LONGITUDE then gps LATITUDE
#' @details
#' You need either package 'rgdal' [need libgdal1-dev in ubuntu and libproj-dev] OR 'proj' in your path [proj-bin in ubuntu]
#' @family coordinates
#' @export
UTM2GPS <- function(utm, ...) {
    return(GPS2UTM(utm, inverse=TRUE, ...))
}

#' formats an nx2 array with `lon, lat` to "\%.10fE, \%.10fS".
#'
#' @inheritParams GPS2UTM
#' @param verbose whether to print the output to screen.
#' @return a single string containing the lon, lat in format
#'         "\%.10fE, \%.10fS", each coordinate separted by a new line.
#' @family coordinates
#' @export
formatGPS <- function(gps, verbose=TRUE) {
    str <- paste(apply(gps, 1, function(x)sprintf("%.10fE, %.10fS", x[1], -x[2])), collapse='\n')
    if (verbose) message(str)
    return(str)
}

#' asks a question and waits for valid input.
#'
#' This is a bit like \code{\link[utils]{menu}} except it doesn't display
#' the choices as \code{1: xx 2: xx} etc. Suitable for yes/no questions
#' where 'menu' is a bit too much.
#'
#' @param question the question to ask the user
#' @param ... the list of valid inputs to chose from, of the form
#' \code{input=action}. This is the same sort of input that should be fed
#' into \code{\link[base]{switch}}. All arguments must be names, except
#' possibly one to be used as the default value (which must occur last).
#' See details.
#' @param case.sensitive whether to coerce all replies to \code{question}
#' to lower case before performing comparison.
#' @param extraArgs a *list* of extra arguments to pass to the function
#' that gets chosen from \code{...}. The format of the list is as for
#' \code{\link[base]{do.call}}.
#' @return the value of one of the elements of \code{...}.
#'
#' @details
#' The answer to \code{question} is collected as a string using
#' \code{\link[base]{readline}}.
#'
#' This is then fed (in string form) to \code{\link[base]{switch}} with the
#' \code{...} arguments: \code{switch(readline(question), ...)}.
#'
#' As such, the default argument (if any) must go last.
#'
#' The argument that is selected is:
#' \itemize{
#'     \item evaluated via \code{do.call(arg, extraArgs)}, if it is a
#'     function.
#'     \item simply returned, otherwise.
#' }
#' The output of this is returned as the output of the function.
#'
#' @examples
#' \dontrun{
#' # example of setting a parameter via a yes/no question:
#' p <- 1
#' p <- verifyInput(sprintf('current value of p is %i, keep? [Y/N] ', p),
#'                   Y=, y=p,                    # return p if they keep
#'                   N=, n= function() readline('Enter new value for p: '))
#' # If the user says 'N' they enter a new value for p and this is
#' # returned.
#' }
#' @export
verifyInput <- function(question, ..., case.sensitive=TRUE, extraArgs=list()) {
    while (TRUE) {
        ans <- readline(question)
        if (!case.sensitive) ans <- tolower(ans)
        fun <- switch(ans, ...)
        if (!is.null(fun)) {
            if (is.function(fun)) {
                return(do.call(fun, extraArgs))
            }
            return(fun)
        }
    }
}


# ************** DEPRECATED ******************* #
####################################
#collectInputVector <- function(question, cases, vectorType="character", case.sensitive=FALSE)
####################################
## Will ask [question] until a 'finished <- TRUE' is evaluated. Amalgamates responses in a vector.
## Input is compared against [cases], a list (e.g. cases=list('Y'='doSomething', 'N'='doSomethingElse', 'default'='somethingElseAgain'))
## The names of the list are compared to the input and if they match the value is eval'd in the FUNCTION (not parent) ENVIRONMENT.
## In order for this function to return you need a 'finished <- TRUE' in at least one of your cases
##  (once this is encountered the function will return all the responses thus far)
## You can have a case 'default' in cases which MUST BE LAST.
#{
#    v        <- vector(mode=vectorType)
#    finished <- FALSE
#    while (!finished) {
#        dmy <- readline(question)
#        v[length(v)+1] <- as(dmy, vectorType)
#        if (!case.sensitive) { dmy <- toupper(dmy) }
#
#        # loop through all other possibilities first
#        for (response in names(cases)) {
#            if (dmy == response) {
#                eval(parse(text=cases[[response]]))
#                break
#            } else if (response == "default") {
#                eval(parse(text=cases[[response]]))
#                break
#            }
#        }
#    } # while (!finished)
#
#    return(v)
#} # collectInput
