# NOTE: this file depends on 'test-includes.r', sourced in utilitiesR/tests/test-all.r
# source('test-includes.r')
context("file utilities")

describe('getProgram', {
    test("exists and is a function", {
        expect_that('getProgram', is_exported('function'))
    })

    test("returns the path to existing programs", {
        # well, R is installed
        R <- getProgram(c('R', 'R.exe'))
        R2 <- Sys.which(c('R', 'R.exe'))
        expect_that(R %in% R2, is_true())
    })

    test("returns NULL for a non-existent program", {
        expect_that(getProgram('!l234lkjsd89\\)X'), is_null())
    })

    # hmm: if you getProgram(c(exists, doesntexist, exists))
    # you will get back a vector of length 2 and you don't necessarily know which
    # one didn't exist.
    # TODO: update the documentation with the above?
    # (I rely on this to return just one value in imageUtilities don't I?)
    # disabled for now.
    # WELL, why do you use this and not Sys.which? it is clearly for a different purpose....
    # so perhaps this is *not* what I want?
    xtest("returns a vector of same length as the input, NA if program not found", {
        rprog <- ifelse(.Platform$OS.type == 'unix', 'R', 'R.exe')
        R <- getProgram(c(rprog, 'alk*32kj)', rprog))
        R2 <- unname(Sys.which(rprog))
        expect_that(R, equals(c(R2, NA, R2)))
    })
})

describe('getExtension', {
    test("exists and is a function", {
        expect_that('getExtension', is_exported('function'))
    })

    test("retrieves the file extension, being the text from the last '.' to the end", {
        expect_that(getExtension('foo.tar'), is_identical_to('tar'))
        expect_that(getExtension('a/foo.tar.gz'), is_identical_to('gz'))
    })

    test("returns '' if there is no extension or filename", {
        expect_that(getExtension('a/b/c/foo'), is_identical_to(''))
        expect_that(getExtension('f.d/a/foo'), is_identical_to(''))
        expect_that(getExtension('f.d/a/'), is_identical_to(''))
    })

    test("is not tricked by hidden UNIX files (start with '.')", {
        expect_that(getExtension('.foo'), is_identical_to(''))
        expect_that(getExtension('/home/user/.bashrc'), is_identical_to(''))
        expect_that(getExtension('/home/user/.bashrc.txt'), is_identical_to('txt'))
    })

    test("is vectorised", {
        expect_that(getExtension(c('a/b/c', 'c.tar', 'foo/bar.tar.gz')), is_identical_to(c('', 'tar', 'gz')))
    })

    test("respects its `preserveCase` argument", {
        expect_that(getExtension('a/b/c.FOO', preserveCase=T), is_identical_to('FOO'))
        expect_that(getExtension('a/b/c.FOO', preserveCase=F), is_identical_to('foo'))
    })
})

describe('getFilename', {
    test("exists and is a function", {
        expect_that('getFilename', is_exported('function'))
    })

    test("retrieves the file name not including extension", {
        # extension is just from the last '.' onwards.
        expect_that(getFilename('foo.tar'), is_identical_to('foo'))
        expect_that(getFilename('a/foo.tar.gz'), is_identical_to('foo.tar'))
        expect_that(getFilename('foo'), is_identical_to('foo'))
        expect_that(getFilename('a.b/foo'), is_identical_to('foo'))
        expect_that(getFilename('a.b/Foo.txt'), is_identical_to('Foo'))
    })

    test("returns '' if there is no filename", {
        expect_that(getFilename('f.d/a/'), is_identical_to(''))
    })

    test("is not tricked by hidden UNIX files (start with '.')", {
        expect_that(getFilename('.foo'), is_identical_to('.foo'))
        expect_that(getFilename('/home/user/.bashrc'), is_identical_to('.bashrc'))
    })

    test("is vectorised", {
        expect_that(getFilename(c('a/b.f/c.exe', 'foo', 'b/x.tar')), is_identical_to(c('c', 'foo', 'x')))
    })
})

describe('createFilename', {
    test("exists and is a function", {
        expect_that('createFilename', is_exported('function'))
    })

    test("performs substitutions", {
        expect_that(createFilename('%f_asdf.jpg', '/tmp/foo/bar.tar.gz'),
                    is_identical_to('bar.tar_asdf.jpg'))
        expect_that(createFilename('%B/%f_%e_%d.jpg', '/tmp/foo/bar.txt', base='FOO/BAR'),
                    is_identical_to('FOO/BAR/bar_txt_foo.jpg'))
    })

    test("can accept extra substitutions in the ... argument", {
        expect_that(createFilename('%f%a%b.png', 'foobar.txt',
                                   extraTokens=list(a='_', b=function(fpath, base) { nchar(fpath) })),
                    is_identical_to('foobar_10.png'))
    })
})

describe('getUniqueFilename', {
    test("exists and is a function", {
        expect_that('getUniqueFilename', is_exported('function'))
    })

    test("returns the input filename if overwrite=T", {
        fn <- 'flk908a7803*O&*#$&*;/fga.k*('
        expect_that(getUniqueFilename(fn, overwrite=T), is_identical_to(fn))
        fn <- 'foo/bar.gz'
        expect_that(getUniqueFilename(fn, overwrite=T), is_identical_to(fn))
    })

    test("returns the input filename if it doesn't exist and otherwise adds numbers on the end until this occurs", {
        pth <- tempfile(pattern='getUniqueFilename')
        while (file.exists(pth)) {
            pth <- tempfile(pattern='getUniqueFilename')
        }
        # pth
        expect_that(getUniqueFilename(paste0(pth, '.txt')),
                    is_identical_to(paste0(pth, '.txt')))
        cat('', file=paste0(pth, '.txt'))

        # pth exists, pth1.txt is the next
        expect_that(out <- getUniqueFilename(paste0(pth, '.txt')),
                    gives_warning('exists already'))
        expect_that(out,
                    is_identical_to(paste0(pth, '1.txt')))

        # also test with no extension
        expect_that(getUniqueFilename(pth), is_identical_to(pth))
        cat('', file=pth)
        # pth exists, pth1.txt is the next
        expect_that(out <- getUniqueFilename(pth),
                    gives_warning('exists already'))
        expect_that(out,
                    is_identical_to(paste0(pth, '1')))
    })

    # TODO: is this a feature I want? (not currently vectorised)
    xtest("is vectorized", {
    })
})

describe('makeIfNotExists', {
    makeNonexistentDir <<- function () {
        # gets the name of a directory under the temp dir that doesn't exist
        pth <- tempfile(pattern='dir')
        while (file.exists(pth)) {
            pth <- tempfile(pattern='dir')
        }
        pth
    }

    test("exists and is a function", {
        expect_that('makeIfNotExists', is_exported('function'))
    })

    test("makes the directory if it does not exist, respecting the `recursive` option", {
        d <- makeNonexistentDir()
        makeIfNotExists(d, recursive=F)
        expect_that(d, is_directory())

        d <- file.path(makeNonexistentDir(), 'subdir')
        makeIfNotExists(d, recursive=T)
        expect_that(d, is_directory())
        expect_that(dirname(d), is_directory())

        # will give a warning (can't make nested directories with recursive=F)
        d <- file.path(makeNonexistentDir(), 'subdir')
        expect_that(makeIfNotExists(d, recursive=F),
                    gives_warning('No such file or directory'))
        expect_that(d, not(file_exists()))
        expect_that(dirname(d), not(file_exists()))
    })

    test("leaves the directory alone if it already exists", {
        # make a directory.
        d <- makeNonexistentDir()
        makeIfNotExists(d)
        expect_that(d, is_directory())
        i <- file.info(d)

        makeIfNotExists(d)
        i2 <- file.info(d)

        # access/modify time is the same
        expect_that(i, is_identical_to(i2))
    })

    test("is vectorized", {
        ds <- c(makeNonexistentDir(), makeNonexistentDir())
        makeIfNotExists(ds)

        expect_that(ds[1], is_directory())
        expect_that(ds[2], is_directory())
    })

    test("prints a message() if the directory does not exist", {
        expect_that(makeIfNotExists(makeNonexistentDir()), shows_message())
        expect_that(suppressMessages(makeIfNotExists(makeNonexistentDir())), not(shows_message()))
    })
})
